@@ -24,6 +24,43 @@
 #' between \eqn{x_{i-10}} and \eqn{x_{i+10}}, resulting in \code{10*2 + 1 = 21} observations falling within
 #' the window.
 #'
+#' @section Parallel Computing:
+#'
+#' Please note that running tasks in parallel does \strong{not} guarantee faster computations.
+#' The overhead introduced is sometimes too large, and it's better to run tasks sequentially.
+#'
+#' The user can register a parallel backend with the \code{doParallel} package in order to attempt to
+#' speed up the calculations (see the examples).
+#'
+#' @examples
+#'
+#' \dontrun{
+#' #### Running TADPole with parallel support
+#' require(doParallel)
+#'
+#' # Load data
+#' data(uciCT)
+#'
+#' # Reinterpolate to same length and coerce as matrix
+#' data <- t(sapply(CharTraj, reinterpolate, newLength = 180))
+#'
+#' # Create parallel workers
+#' cl <- makeCluster(detectCores())
+#' registerDoParallel(cl)
+#'
+#' # Cluster
+#' kc.tadp <- TADPole(data, k = 20, window.size = 20, dc = 1.5)
+#'
+#' # Stop parallel workers
+#' stopCluster(cl)
+#'
+#' # Return to sequential computations
+#' registerDoSEQ()
+#'
+#' # Compute Rand Index
+#' cat("Rand index for TADPole:", randIndex(kc.tadp$cl, CharTrajLabels), "\n\n")
+#' }
+#'
 #' @references
 #'
 #' Begum N, Ulanova L, Wang J and Keogh E (2015). ``Accelerating Dynamic Time Warping Clustering with a Novel Admissible Pruning
@@ -42,10 +79,18 @@
 #'   \item \code{distCalcPercentage}: Percentage of distance calculations that were actually performed.
 #' }
 #'
+#' @import foreach
+#' @importFrom parallel splitIndices
+#'
 #' @export
 
 TADPole <- function(data, window.size = NULL, k = 2, dc, error.check = TRUE) {
 
+     if (missing(window.size))
+          stop("Please provide a positive window size")
+     if (missing(dc))
+          stop("Please provide the 'dc' parameter")
+
      x <- consistency_check(data, "tsmat")
 
      n <- length(x)
@@ -63,18 +108,27 @@ TADPole <- function(data, window.size = NULL, k = 2, dc, error.check = TRUE) {
 
      UBM <- proxy::dist(x, x, method = "L2")
 
+     ## Euclidean is only valid as upper bound if 'symmetric1' step pattern is used
      step.pattern <- get("symmetric1") # so that CHECK doesn't complain
 
+     ## Attempt parallel computations?
+     do_par <- foreach::getDoParRegistered()
+
+     if (do_par)
+          workers <- foreach::getDoParWorkers()
+     else
+          workers <- 1L
+
      ## ============================================================================================================================
      ## Pruning during local density calculation
      ## ============================================================================================================================
 
      ## Flag definition
-     # 0 # DTW calculated, and it lies below dc
-     # 1 # calculate DTW
-     # 2 # within dc, prune
-     # 3 # not within dc, prune
-     # 4 # identical series
+     # 0 - DTW calculated, and it lies below dc
+     # 1 - calculate DTW
+     # 2 - within dc, prune
+     # 3 - not within dc, prune
+     # 4 - identical series
 
      ## Initialize matrices
      D <- matrix(NA, n, n)
@@ -105,10 +159,33 @@ TADPole <- function(data, window.size = NULL, k = 2, dc, error.check = TRUE) {
      ind1 <- which(Flags == 1, arr.ind = TRUE)
 
      if (nrow(ind1) > 0) {
-          d1 <- proxy::dist(x[ind1[, 1]], x[ind1[, 2]], method = "DTW2",
-                            step.pattern = step.pattern,
-                            window.type = "slantedband", window.size = window.size,
-                            pairwise = TRUE)
+
+          ## Attempt parallel calculations?
+          if (do_par) {
+               tasks <- parallel::splitIndices(nrow(ind1), workers)
+
+               ind1 <- lapply(tasks, function(id) {
+                    ind1[id,]
+               })
+
+               exclude <- setdiff(ls(), c("x", "step.pattern", "window.size"))
+
+               d1 <- foreach(ind1 = ind1,
+                             .combine = c,
+                             .multicombine = TRUE,
+                             .packages = "dtwclust",
+                             .noexport = exclude) %dopar% {
+                                  proxy::dist(x[ind1[, 1]], x[ind1[, 2]], method = "DTW2",
+                                              step.pattern = step.pattern,
+                                              window.type = "slantedband", window.size = window.size,
+                                              pairwise = TRUE)
+                             }
+          } else {
+               d1 <- proxy::dist(x[ind1[, 1]], x[ind1[, 2]], method = "DTW2",
+                                 step.pattern = step.pattern,
+                                 window.type = "slantedband", window.size = window.size,
+                                 pairwise = TRUE)
+          }
 
           ## Fill distance matrix where necessary
           D[which(Flags==1)] <- d1
@@ -163,35 +240,82 @@ TADPole <- function(data, window.size = NULL, k = 2, dc, error.check = TRUE) {
      ## Pruning during NN distance calculation from higher density list (phase 2)
      ## ============================================================================================================================
 
-     DNN <- t(sapply(2:n, function (i) {
-          ## Index of higher density neighbors
-          indHDN <- TADPorder$ix[1:(i-1)]
-          ## Index of current object
-          ii <- TADPorder$ix[i]
+     ## Attempt parallel calculations?
+     if (do_par) {
+          tasks <- parallel::splitIndices(n-1L, workers)
 
-          ## If this is true, prune the calculation of the true distance
-          indPrune <- LBM[ii, indHDN] > deltaUB[ii]
-          ## If the distance was already computed before, don't do it again
-          indPre <- Flags[ii, indHDN] == 0 | Flags[ii, indHDN] == 1
+          i <- lapply(tasks, function(id) {
+               id + 1 # start at two
+          })
 
-          ## 'delta' will have the distances to neighbors with higher densities. Initially filled with upper bound
-          delta <- UBM[ii, indHDN]
-          ## If some distances were already computed, put them here
-          delta[indPre] <- D[ii, indHDN[indPre]]
+          DNN <- foreach(i = i,
+                         .combine = rbind,
+                         .multicombine = TRUE,
+                         .packages = "dtwclust") %dopar% {
 
-          ## If the distance is not to be pruned nor previously calculated, compute it now
-          indCompute <- !(indPrune | indPre)
+                              t(sapply(i, function (i) {
+                                   ## Index of higher density neighbors
+                                   indHDN <- TADPorder$ix[1:(i-1)]
+                                   ## Index of current object
+                                   ii <- TADPorder$ix[i]
 
-          if (sum(indCompute) > 0) {
-               d2 <- proxy::dist(x[ii], x[indHDN[indCompute]], method = "DTW2",
-                                 step.pattern = step.pattern,
-                                 window.type = "slantedband", window.size = window.size)
+                                   ## If this is true, prune the calculation of the true distance
+                                   indPrune <- LBM[ii, indHDN] > deltaUB[ii]
+                                   ## If the distance was already computed before, don't do it again
+                                   indPre <- Flags[ii, indHDN] == 0 | Flags[ii, indHDN] == 1
 
-               delta[indCompute] <- d2
-          }
+                                   ## 'delta' will have the distances to neighbors with higher densities. Initially filled with upper bound
+                                   delta <- UBM[ii, indHDN]
+                                   ## If some distances were already computed, put them here
+                                   delta[indPre] <- D[ii, indHDN[indPre]]
 
-          c(delta = min(delta), NN = indHDN[which.min(delta)], distCalc = sum(indCompute))
-     }))
+                                   ## If the distance is not to be pruned nor previously calculated, compute it now
+                                   indCompute <- !(indPrune | indPre)
+
+                                   if (sum(indCompute) > 0) {
+                                        d2 <- proxy::dist(x[ii], x[indHDN[indCompute]], method = "DTW2",
+                                                          step.pattern = step.pattern,
+                                                          window.type = "slantedband", window.size = window.size)
+
+                                        delta[indCompute] <- d2
+                                   }
+
+                                   c(delta = min(delta), NN = indHDN[which.min(delta)], distCalc = sum(indCompute))
+                              }))
+                         }
+
+     } else {
+          DNN <- t(sapply(2:n, function (i) {
+               ## Index of higher density neighbors
+               indHDN <- TADPorder$ix[1:(i-1)]
+               ## Index of current object
+               ii <- TADPorder$ix[i]
+
+               ## If this is true, prune the calculation of the true distance
+               indPrune <- LBM[ii, indHDN] > deltaUB[ii]
+               ## If the distance was already computed before, don't do it again
+               indPre <- Flags[ii, indHDN] == 0 | Flags[ii, indHDN] == 1
+
+               ## 'delta' will have the distances to neighbors with higher densities. Initially filled with upper bound
+               delta <- UBM[ii, indHDN]
+               ## If some distances were already computed, put them here
+               delta[indPre] <- D[ii, indHDN[indPre]]
+
+               ## If the distance is not to be pruned nor previously calculated, compute it now
+               indCompute <- !(indPrune | indPre)
+
+               if (sum(indCompute) > 0) {
+                    ## TODO: parallel support
+                    d2 <- proxy::dist(x[ii], x[indHDN[indCompute]], method = "DTW2",
+                                      step.pattern = step.pattern,
+                                      window.type = "slantedband", window.size = window.size)
+
+                    delta[indCompute] <- d2
+               }
+
+               c(delta = min(delta), NN = indHDN[which.min(delta)], distCalc = sum(indCompute))
+          }))
+     }
 
      ## To order according to default order
      indOrig <- sort(TADPorder$ix, index.return = TRUE)$ix
